/*  SPDX-License-Identifier: LGPL-2.0-or-later

    SPDX-FileCopyrightText: 2019 Christoph Cullmann <cullmann@kde.org>

    SPDX-License-Identifier: LGPL-2.0-or-later
*/

#ifndef KATE_DEBUG_H
#define KATE_DEBUG_H

#include <QLoggingCategory>
Q_DECLARE_LOGGING_CATEGORY(LOG_KATE)

#endif /* KATE_DEBUG_H */
